import argparse
import unittest
import warnings
import os
from update_mapsource.update_mapsource import update_mapsource
from lxml import etree as ET


def ignore_warnings(test_func):
    def do_test(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            test_func(self, *args, **kwargs)

    return do_test


class Test_update_mapsource(unittest.TestCase):

    THIS_DIR = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))

    def setUp(self):
        self.data_path = os.path.normpath(
            os.path.join(self.THIS_DIR, os.pardir, 'tests/mapfiles/'))
        # print("Data path: %s" % self.data_path)

    @ignore_warnings
    def test_uniq(self):
        u = update_mapsource()

        parser = argparse.ArgumentParser()

        parser.add_argument('--sourcefile')
        parser.add_argument('--mapfile')
        parser.add_argument('--outputfile')
        parser.add_argument('--datasource')
        parser.add_argument('--loglevel')

        args = parser.parse_args(args=['--sourcefile', f'{self.data_path}/dmso.xml',
                                       '--mapfile', f'{self.data_path}/testuniq.map',
                                       '--datasource', f'{self.data_path}/dmso.xml',
                                       '--outputfile', '/tmp/testuniq.xml', '--loglevel=DEBUG'
                                       ])

        print(args.sourcefile)
        u.clean_mapfile(args.mapfile)
        u.process(args)

        tree = ET.parse("/tmp/testuniq.xml")
        uniq = tree.find(".//UniqueField")
        self.assertTrue(uniq is not None)
        self.assertEquals('ogc_fid', uniq.text)

    @ignore_warnings
    def test_read_map_file(self):

        for file in ("dmso", "standard"):
            u = update_mapsource()

            parser = argparse.ArgumentParser()

            parser.add_argument('--sourcefile')
            parser.add_argument('--mapfile')
            parser.add_argument('--outputfile')
            parser.add_argument('--datasource')

            args = parser.parse_args(args=['--sourcefile', f'{self.data_path}/{file}.xml',
                                           '--mapfile', f'{self.data_path}/{file}.map',
                                           '--datasource', f'{self.data_path}/{file}.xml',
                                           '--outputfile', f'/tmp/test_{file}.out'])

            u.clean_mapfile(args.mapfile)
            u.process(args)

            parser = ET.XMLParser(remove_blank_text=True)
            tree = ET.parse(f"/tmp/test_{file}.out", parser)
            self.assertLess(7, len(tree.xpath(".//Type[text()='ows']")))

    @ignore_warnings
    def test_null_data(self):

        file = 'null'
        u = update_mapsource()

        parser = argparse.ArgumentParser()

        parser.add_argument('--sourcefile')
        parser.add_argument('--mapfile')
        parser.add_argument('--outputfile')
        parser.add_argument('--datasource')

        args = parser.parse_args(args=['--sourcefile', f'{self.data_path}/{file}.xml',
                                       '--mapfile', f'{self.data_path}/{file}.map',
                                       '--datasource', f'{self.data_path}/{file}.xml',
                                       '--outputfile', f'/tmp/test_{file}.out'])

        u.clean_mapfile(args.mapfile)
        u.process(args)

        parser = ET.XMLParser(remove_blank_text=True)
        tree = ET.parse(f"/tmp/test_{file}.out", parser)
        sql = tree.xpath(".//Data/SQL")[0]
        print(ET.dump(sql))
        self.assertTrue(sql.text.endswith("labeltext is null"))

    @ignore_warnings
    def test_connection(self):

        file = 'connection'
        u = update_mapsource()

        parser = argparse.ArgumentParser()

        parser.add_argument('--sourcefile')
        parser.add_argument('--mapfile')
        parser.add_argument('--outputfile')
        parser.add_argument('--datasource')

        args = parser.parse_args(args=['--sourcefile', f'{self.data_path}/{file}.xml',
                                       '--mapfile', f'{self.data_path}/{file}.map',
                                       '--datasource', f'{self.data_path}/workflow.xml',
                                       '--outputfile', f'/tmp/test_{file}.out'])

        u.clean_mapfile(args.mapfile)
        u.process(args)

        parser = ET.XMLParser(remove_blank_text=True)
        tree = ET.parse(f"/tmp/test_{file}.out", parser)
        sql = tree.xpath(".//Connection")[0]
        print(ET.dump(sql))
        self.assertTrue(sql.text is None)
        self.assertEqual("SDW", sql.attrib['name'])
