'''
Created on 15 Feb 2017

@author: ian
'''
import logging
from maputils import xml_to_sld
from maputils import map_to_xml
import io
import shutil
import argparse
import lxml.etree as ET
from lxml.etree import CDATA
from pathlib import Path, PureWindowsPath
import regex as re
import mappyfile


class update_mapsource:
    datasources = []
    replacements = {}

    def clean_mapfile(self, mapfile):
        backup = mapfile + '.bak'
        shutil.copy(mapfile, mapfile + ".bak")
        lines = []
        with open(backup, 'rt') as f:
            while True:
                line = f.readline()
                if not line:
                    break
                replaced = line
                if line.strip().upper().startswith('DATA'):
                    pattern = pattern = r'\\"+\w+\\"'
                    matches = re.findall(pattern, line)
                    if matches:
                        for ma in matches:
                            self.replacements[ma[2:-2]] = ma
                    replaced = line.replace('\\"', '')
                else:
                    replaced = line.replace('\\', '/')
                lines.append(replaced)
            # save file
        with open(backup, 'wt') as f:
            for line in lines:
                f. write(line)
        return backup

    def parse_xmlfile(self, input_file):
        parser = ET.XMLParser(remove_blank_text=True)
        tree = ET.parse(input_file, parser)
        return tree.getroot()

    def process(self, options):

        logging.info("updating " + options.sourcefile + " with " + options.mapfile)
        if options.outputfile:
            logging.info("results in " + options.outputfile)
        mapfile = self.clean_mapfile(options.mapfile)
        mapf = map_to_xml.map_to_xml(mapfile, expand_includes=True)
        # an unexpanded version of the mapfile for output.
        # parse the mapfile into an XML tree
        with io.open(options.mapfile, "r", encoding="utf-8") as mpfile:
            self.mf = mappyfile.load(mpfile, expand_includes=True)

        # Convert the mapfile to a collection of SLD keyed by layer name
        self.sldStore = xml_to_sld.xml_to_sld("", root=mapf.map_root)
        # extract the dict
        layer_info = self.sldStore.layer_info
        logging.info(f"processing {len(layer_info)} layers")

        # Parse the existing MapSource file
        self.mapsource = self.parse_xmlfile(options.sourcefile)

        self.datasources = self.load_datasource(options.datasource)
        processed_layers = set()
        # for each layer in the MapSource
        for source_layer in self.mapsource.iter('Layer'):

            rows = source_layer.find("Rows")
            if rows is not None:
                # we need to grab the contents of the row elements
                # and wrap them in CDATA
                for row in rows:
                    row.text = CDATA(row.text)

            name = source_layer.find("Name")
            if name is None:
                continue
                # look up the layer name in the MapSource
            tname = source_layer.find('LayerName')
            # type = source_layer.find('Type').text
            if tname is not None and tname.text in layer_info:
                # the layer is found in the MAP file?
                layer = layer_info[tname.text]
            else:
                logging.info("'" + name.text + "' is already an OGC layer")
                continue
            if tname.text in processed_layers:
                logging.warning(f"{tname.text} is repeated - skipping this time")
                source_layer.getparent().remove(source_layer)
                continue
            processed_layers.add(tname.text)
            logging.debug("Processing " + name.text + " (" + tname.text + ")")

            ows = source_layer.find("OWS")
            # is it a classic layer?
            if ows is None:
                self.process_classic_layer(source_layer, layer)

        # write out new mapfile
        base = Path(options.outputfile).parent
        basename = Path(options.outputfile).stem
        outfile = Path.joinpath(base, basename + '.map')
        logging.debug(f"Saving mapfile to {outfile}")
        logging.debug(f"checking if {outfile.parent} exists")
        if not outfile.parent.exists():
            logging.debug(f"creating {outfile.parent}")
            Path(outfile.parent).mkdir(parents=True, exist_ok=True)
        mappyfile.save(self.mf, outfile)
        # update the map file name in the output
        name = self.mapsource.find("MapName")
        if name is not None:
            old = PureWindowsPath(name.text)
            old = old.parent
            basename = Path(options.outputfile).stem + '.map'
            pretendFile = Path.joinpath(old, Path(basename))
            name.text = str(pretendFile)

            xml_str = ET.tostring(self.mapsource,
                                  pretty_print=True,
                                  xml_declaration=True,
                                  encoding='UTF-8')
        if options.outputfile:
            with open(options.outputfile, 'wb') as o:
                o.write(xml_str)
        else:
            print(xml_str)

    def find_parens(self, s):
        """
        Search for matching parens - need this to extract select from data string
        """
        toret = {}
        pstack = []

        for i, c in enumerate(s):
            if c == '(':
                pstack.append(i)
            elif c == ')':
                if len(pstack) == 0:
                    raise IndexError("No matching closing parens at: " + str(i))
                toret[pstack.pop()] = i

        if len(pstack) > 0:
            raise IndexError("No matching opening parens at: " + str(pstack.pop()))

        return toret

    def get_simple_schema_and_table(self, s):
        """Returns the schema and table from a simple
        SQL statement, otherwise None"""

        # replace groups of whitespace with a single space
        s = re.sub(r'\s{2,}', ' ', s)

        # tokenise the string
        parts = s.split(' ')

        if len(parts) != 4:
            return False

        if (parts[0].lower() != 'select' or parts[1] != '*' or parts[2].lower() != 'from'):
            return False

        if re.search(r'(\"?[^\"\.]+\"?\.)?(\"?[^\"]+\"?)', s[3]) is None:
            return False

        return parts[3]

    def load_datasource(self, datasource):
        logging.info("loading workflow " + datasource)
        data = self.parse_xmlfile(datasource)
        # lookup some data sources?
        datasources = {}
        for connections in data.iter('Connections'):
            for conn in connections.iter("Connection"):
                name = conn.find("Name")
                db = conn.find("Database")
                if name is not None and db is not None \
                        and db.text not in datasources:
                    # print db.text + "=>" + name.text
                    datasources[db.text] = name.text
        return datasources

    def process_classic_layer(self, source_layer, layer):
        t = source_layer.find("Type")
        tname = source_layer.find('LayerName')
        if t is None:
            source_layer.insert(
                source_layer.index(tname) + 1, ET.XML("<Type>ows</Type>"))
        else:
            t.text = "ows"
        ows = ET.SubElement(source_layer, "OWS")
        geom = ET.SubElement(ows, "Geometry")
        type = ET.SubElement(geom, "Type")
        type.text = layer.attrib['type']
        # at present there is no way to determine
        # if this is a multi geom
        type.attrib['multi'] = "false"
        source = ET.SubElement(geom, "Source")  # no idea what goes here
        source.text = ""
        field = ET.SubElement(geom, "Field")
        # extract the DATA layer - we need the select statement.
        ds = layer.find("{http://www.mapserver.org/mapserver}Data")
        if ds is not None:
            # find the first word (that is the column name
            # we want in Field)
            index = ds.text.find(' ')
            data = ds.text[:index]
            field.text = data
        sld = ET.SubElement(ows, "SLD")
        fts = self.sldStore.getLayer(tname.text)

        # logging.debug(f"{tname.text} finds {fts} as SLD {sld}")
        sld.append(fts)
        data = ET.SubElement(ows, "Data")
        conn = ET.SubElement(data, "Connection")
        logging.debug(ET.dump(layer))
        c = layer.find("{http://www.mapserver.org/mapserver}connection")

        if c is not None:
            parts = c.text.split()
            logging.debug(f"connection: {c.text}")
            for p in parts:
                logging.debug(f"looking at part {p} for dbname")
                bits = p.split("=")
                if bits[0] == 'dbname' and bits[1] in self.datasources:
                    conn.set("name", self.datasources[bits[1]])
                    logging.debug(f"connection set to {self.datasources[bits[1]]} for {layer.attrib['name']}")
                else:
                    if bits[0] == 'dbname':
                        logging.info(f"Can't find {bits[1]} in datasources {self.datasources}")
        else:
            logging.debug("No connection found for " + layer.attrib['name'])
        sql = ET.SubElement(data, "SQL")
        fields = source_layer.find("Fields")
        if fields is None:
            fields = ET.SubElement(layer, "Fields")

        names = set([x.text.strip("][") for x in fields.findall(".//Field") if x.text is not None])
        properties = set([x.text.strip("][") for x in sld.findall(".//PropertyName") if x.text is not None])
        labels = set([x.text.strip("][") for x in sld.findall(".//Label") if x.text is not None])
        properties = labels | properties
        for x in properties.difference(names):
            f = ET.SubElement(fields, "Field")
            f.text = x
            f.attrib['Type'] = 'AttributeRule'
            f.attrib['TextChange'] = 'false'

        uniq = ET.SubElement(data, "UniqueField")
        if ds is not None:
            # now we want the contents of the select statement.
            # find the opening (
            p = self.find_parens(ds.text)
            start = -1
            end = len(ds.text)
            for key in sorted(p.keys()):
                start = key
                end = p[key]
                break
            sql_text = ds.text[start + 1:end]
            # check to see if it is a simple SQL statement
            # (SELECT * FROM "schema"."table") and if so,
            # use just the schema and table as the SQL for
            # the mapsource
            sql_table = self.get_simple_schema_and_table(sql_text)
            if sql_table is not None and sql_table:
                sql_text = sql_table

            for mat in self.replacements.keys():
                if mat in sql_text:
                    sql_text = sql_text.replace(mat, self.replacements.get(mat))
            if "'" in sql_text:
                logging.debug(f"replacing ' in {sql_text}")
                sql_text = sql_text.replace("'", "$$")
            sql.text = CDATA(str(sql_text))
            key = 'using unique '
            index = ds.text.find(key)
            start = index + len(key)
            end = ds.text.find(' ', start)
            if end < 0:
                uniq.text = ds.text[start:]
            else:
                uniq.text = ds.text[start:end]
        else:
            logging.warn("No Data statement found")
        # now remove that layer from the output mapfile.
        if self.mf:
            ml = mappyfile.find(self.mf["layers"], "name", tname.text)
            if ml:
                logging.debug("removing " + ml['name'] + " from output mapfile")
                self.mf["layers"].remove(ml)


def main():

    parser = argparse.ArgumentParser(prog='update_mapsource')
    parser.add_argument("-s",
                        "--sourcefile",
                        dest="sourcefile",
                        help="iShare source file",
                        metavar="SOURCEFILE.xml",
                        required=True)
    parser.add_argument("-m",
                        "--mapfile",
                        dest="mapfile",
                        help="MapServer map file",
                        metavar="MAPFILE.map",
                        required=True)
    parser.add_argument("-w",
                        "--workflow",
                        dest="datasource",
                        help="iShare workflow file",
                        metavar="workflow.xml",
                        required=True)
    parser.add_argument("-o",
                        "--output",
                        dest="outputfile",
                        default="",
                        help="new iShare MapSource file",
                        metavar="FILE",
                        required=True)
    parser.add_argument("-l",
                        "--loglevel",
                        dest="loglevel",
                        default='INFO',
                        help='One of DEBUG, WARNING, ERROR, CRITICAL')
    options = parser.parse_args()

    numeric_level = getattr(logging, options.loglevel.upper(), None)

    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % options.loglevel)
    logging.basicConfig(level=numeric_level,
                        format='%(levelname)-8s %(message)s')
    logging.debug(options)
    # copy the mapfile to a backup

    u = update_mapsource()
    u.process(options)


if __name__ == "__main__":
    main()
